//
// topo_gen.cc   : Topology generation program
// author           : Fabio Silva
//
// Copyright (C) 2000-2005 by the University of Southern California
// $Id: topo_gen.cc,v 1.2 2005/05/10 04:12:06 fstann Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//

#include "topo_gen.hh"

double Distance(double x1, double y1, double x2, double y2)
{
  return sqrt((x1 - x2) * (x1 -x2) + (y1 - y2) * (y1 - y2));
}

int TopoGen::findClosest(double x, double y)
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;
  double closest_distance;
  double current_distance;
  int closest;

  closest = -1;
  closest_distance = Distance(0, 0, MAX_X, MAX_Y);

  for (node_iterator = nodes_.begin(); node_iterator != nodes_.end(); ++node_iterator){
    node_entry = *node_iterator;

    current_distance = Distance(x, y,
                 node_entry->node_x_, node_entry->node_y_);

    // Is this node closest to what we have ?
    if (current_distance > closest_distance)
      continue;

    // Yes, keep track of it
    closest_distance = current_distance;
    closest = node_entry->node_id_;
  }
  
  return closest;
}

NodeInfo * TopoGen::findNode(NodeList *nodes, int id)
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;

  for (node_iterator = nodes->begin();
       node_iterator != nodes->end(); ++node_iterator){
    node_entry = *node_iterator;

    if (node_entry->node_id_ == id)
      return node_entry;
  }

  // Node cannot be found !
  return NULL;
}

bool TopoGen::deleteNode(int id)
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;

  for (node_iterator = nodes_.begin(); node_iterator != nodes_.end(); ++node_iterator){
    node_entry = *node_iterator;

    if (node_entry->node_id_ == id){
      // Found it !
      node_iterator = nodes_.erase(node_iterator);
      return true;
    }
  }

  return false;
}

void TopoGen::createNodes()
{
  NodeInfo *node;
  double x, y;

  // Create nodes
  for (int i = 1; i <= number_of_nodes_; i++){
    x = 1.0 * size_x_ * (GetRand() * 1.0 / RAND_MAX);
    y = 1.0 * size_y_ * (GetRand() * 1.0 / RAND_MAX);

    // Add this node to the list
    node = new NodeInfo(i, x, y);
    nodes_.push_back(node);

    // Also add a copy to the topology list
    node = new NodeInfo(i, x, y);
    topology_.push_back(node);
  }
}

bool TopoGen::pickClusteredSinks(double x, double y)
{
  return pickClusteredNodes(&sinks_, number_of_sinks_, x, y);
}

bool TopoGen::pickClusteredSources(double x, double y)
{
  return pickClusteredNodes(&sources_, number_of_sources_, x, y);
}

bool TopoGen::pickClusteredNodes(NodeList *destination_list,
                    int num_nodes, double x, double y)
{
  NodeInfo *node;
  int node_id;

  for (int i = 0; i < num_nodes; i++){
    node_id = findClosest(x, y);

    node = findNode(&nodes_, node_id);

    if (deleteNode(node_id))
      destination_list->push_back(node);
    else
      return false;
  }

  return true;
}

bool TopoGen::pickRandomSinks()
{
  return pickRandomNodes(&sinks_, number_of_sinks_);
}

bool TopoGen::pickRandomSources()
{
  return pickRandomNodes(&sources_, number_of_sources_);
}

bool TopoGen::pickRandomNodes(NodeList *destination_list,
                 int num_nodes)
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;

  // Pick the first num_nodes from nodes_
  for (int i = 0; i < num_nodes; i++){

    node_iterator = nodes_.begin();

    // We must still have nodes available in the nodes_ list
    if (node_iterator == nodes_.end())
      return false;

    node_entry = *node_iterator;

    // Delete node from the nodes_ list
    node_iterator = nodes_.erase(node_iterator);

    // Add node to the destination_list
    destination_list->push_back(node_entry);
  }

  return true;
}

void TopoGen::printConfig()
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;
  char topology_filename[100];
  
  char output_filename[100];
  char custom_output_filename[100];
  FILE *topology_file, *output_file, *custom_output_file;

  // Open output files with user specified name if provided
  if (output_filename_){
    strcpy(output_filename, output_filename_);

    // Create a second file name if we are using a custom filter
    if (use_custom_filter_){
      strcpy(custom_output_filename, output_filename_);
      strcat(custom_output_filename, "-");
      strcat(custom_output_filename, CUSTOM_FILTER_NAME);
    }
  }
  else{
    // Otherwise automatically generate file names
    // Changed to include send rate in the file name
    if (mode_ == TPP){
      if (filenames_with_send_rate_)
        sprintf(output_filename, "topo-tpp-%d-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        send_rate_, number_of_runs_);
      else
        sprintf(output_filename, "topo-tpp-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        number_of_runs_);
    }
    else if (mode_ == OPP){
      if (filenames_with_send_rate_)
        sprintf(output_filename, "topo-opp-%d-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        send_rate_, number_of_runs_);
      else
        sprintf(output_filename, "topo-opp-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        number_of_runs_);
    }
    else{
      if (filenames_with_send_rate_)
        sprintf(output_filename, "topo-push-%d-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        send_rate_, number_of_runs_);
      else
        sprintf(output_filename, "topo-push-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        number_of_runs_);
    }

    // Now concantenate the custom string if use_custom_filter
    // Now add the file prefix
    if (use_custom_filter_){
      strcpy(custom_output_filename, output_filename);
      strcat(custom_output_filename, "-");
      strcat(custom_output_filename, CUSTOM_FILTER_NAME);
    }
  }

  // Open the files
  strcat(output_filename, ".sim");
  output_file = fopen(output_filename, "w");
  if (!output_file){
    fprintf(stderr, "Cannot create %s\n", output_filename);
    exit(-1);
  }
  if (use_custom_filter_){
    strcat(custom_output_filename, ".sim");
    custom_output_file = fopen(custom_output_filename, "w");
    if (!custom_output_file){
      fprintf(stderr, "Cannot create %s\n", custom_output_filename);
      exit(-1);
    }
  }

  // Print Configuration Parameters to the "normal" file
  fprintf(output_file, "# This is an topo_gen generated file\n#\n");
  fprintf(output_file, "# Seed: %ld, Nodes: %d, Sinks: %d, Sources: %d\n",
      seed_, number_of_nodes_, number_of_sinks_, number_of_sources_);
  fprintf(output_file, "# Simulation Area: %d x %d, Send rate: %d\n",
      size_x_, size_y_, send_rate_);
  fprintf(output_file, "# Mode: %d, Sinks(%.2f, %.2f), Sources(%.2f, %.2f)\n",
      mode_, a_x_, a_y_, b_x_, b_y_);

  // First we start with the top level topo_gen stuff
  fprintf(output_file, "\n#Number of nodes\nnum-nodes = %d;\n\n",
      number_of_nodes_);

  fprintf(output_file, "#Field size\nfield-size = (%d, %d);\n\n",
      size_x_, size_y_);

  fprintf(output_file, "#Simulate MoteNIC\n");
  fprintf(output_file, "sim-component = \"sim/sim_radio -m %s -p %d\";\n\n",
    SIM_RADIO_MODEL, SIM_RADIO_POWER);

  fprintf(output_file, "node default {\n");
  fprintf(output_file, "\tpower = on;\n");

  if (mode_ == TPP)
    fprintf(output_file, "\temruntab = " TPP_DEFAULT_TAB ";\n");
  else if (mode_ == OPP)
    fprintf(output_file, "\temruntab = " OPP_DEFAULT_TAB ";\n");
  else
    fprintf(output_file, "\temruntab = " PUSH_DEFAULT_TAB ";\n");

  fprintf(output_file, "}\n\n");

  // Print sinks
  fprintf(output_file, "#Sinks\n");

  for (node_iterator = sinks_.begin();
       node_iterator != sinks_.end(); node_iterator++){
    node_entry = *node_iterator;

    fprintf(output_file, "node %d {\n", node_entry->node_id_);
    fprintf(output_file, "\tposition = (%.2f, %.2f);\n",
        node_entry->node_x_, node_entry->node_y_);
    if (mode_ == TPP)
      fprintf(output_file, "\temruntab = " TPP_SINK_TAB ";\n");
    else if (mode_ == OPP)
      fprintf(output_file, "\temruntab = " OPP_SINK_TAB ";\n");
    else
      fprintf(output_file, "\temruntab = " PUSH_SINK_TAB ";\n");

    fprintf(output_file, "}\n\n");
  }

  // Print sources
  fprintf(output_file, "#Sources\n");

  for (node_iterator = sources_.begin();
       node_iterator != sources_.end(); node_iterator++){
    node_entry = *node_iterator;

    fprintf(output_file, "node %d {\n", node_entry->node_id_);
    fprintf(output_file, "\tposition = (%.2f, %.2f);\n",
        node_entry->node_x_, node_entry->node_y_);
    if (mode_ == TPP){
      if(send_lambda_ > 0)
        fprintf(output_file, "\temruntab = \"" TPP_SOURCE_TAB " %.4f\";\n", send_lambda_);
      else
        fprintf(output_file, "\temruntab = " TPP_SOURCE_TAB ";\n");
    }
    else if (mode_ == OPP){
      if(send_lambda_ > 0)
        fprintf(output_file, "\temruntab = \"" OPP_SOURCE_TAB " %.4f\";\n", send_lambda_);
      else
        fprintf(output_file, "\temruntab = " OPP_SOURCE_TAB ";\n");
    }
    else{
      if(send_lambda_ > 0)
        fprintf(output_file, "\temruntab = \"" PUSH_SOURCE_TAB " %.4f\";\n", send_lambda_);
      else
        fprintf(output_file, "\temruntab = " PUSH_SOURCE_TAB ";\n");
    }

    fprintf(output_file, "}\n\n");
  }

  // Print other nodes
  fprintf(output_file, "#Intermediate Nodes\n");

  for (node_iterator = nodes_.begin();
       node_iterator != nodes_.end(); node_iterator++){
    node_entry = *node_iterator;

    fprintf(output_file, "node %d {\n", node_entry->node_id_);
    fprintf(output_file, "\tposition = (%.2f, %.2f);\n",
        node_entry->node_x_, node_entry->node_y_);

    fprintf(output_file, "}\n\n");
  }

  // Print Stop time
  fprintf(output_file, "time %dm %ds {\n",
      EXPERIMENT_RUN_MINUTES, EXPERIMENT_RUN_SECONDS);
  fprintf(output_file, "\thalt;\n");
  fprintf(output_file, "}\n");

  if (output_file != stdout)
    fclose(output_file);

  /////////////////////////// Custom Filter //////////////////////////
  // Print to the "custom filter" file
  if (use_custom_filter_){
    // Print Configuration Parameters
    fprintf(custom_output_file, "# This is an topo_gen generated file\n#\n");
    fprintf(custom_output_file, "# Seed: %ld, Nodes: %d, Sinks: %d, Sources: %d\n",
    seed_, number_of_nodes_, number_of_sinks_, number_of_sources_);
    fprintf(custom_output_file, "# Simulation Area: %d x %d, Send rate: %d\n",
    size_x_, size_y_, send_rate_);
    fprintf(custom_output_file, "# Mode: %d, Sinks(%.2f, %.2f), Sources(%.2f, %.2f)\n",
    mode_, a_x_, a_y_, b_x_, b_y_);

    // First we start with the top level topo_gen stuff
    fprintf(custom_output_file, "\n#Number of nodes\nnum-nodes = %d;\n\n",
    number_of_nodes_);

    fprintf(custom_output_file, "#Field size\nfield-size = (%d, %d);\n\n",
    size_x_, size_y_);

    fprintf(custom_output_file, "#Simulate MoteNIC\n");
    fprintf(custom_output_file, "sim-component = \"sim/sim_radio -m %s -p %d\";\n\n",
      SIM_RADIO_MODEL, SIM_RADIO_POWER);

    fprintf(custom_output_file, "node default {\n");
    fprintf(custom_output_file, "\tpower = on;\n");

    if (mode_ == TPP)
      fprintf(custom_output_file, "\temruntab = " TPP_CUSTOM_DEFAULT_TAB ";\n");
    else if (mode_ == OPP)
      fprintf(custom_output_file, "\temruntab = " OPP_CUSTOM_DEFAULT_TAB ";\n");
    else
      fprintf(custom_output_file, "\temruntab = " PUSH_CUSTOM_DEFAULT_TAB ";\n");

    fprintf(custom_output_file, "}\n\n");

    // Print sinks
    fprintf(custom_output_file, "#Sinks\n");

    for (node_iterator = sinks_.begin();
    node_iterator != sinks_.end(); node_iterator++){
      node_entry = *node_iterator;

      fprintf(custom_output_file, "node %d {\n", node_entry->node_id_);
      fprintf(custom_output_file, "\tposition = (%.2f, %.2f);\n",
        node_entry->node_x_, node_entry->node_y_);
      if (mode_ == TPP)
        fprintf(custom_output_file, "\temruntab = " TPP_CUSTOM_SINK_TAB ";\n");
      else if (mode_ == OPP)
        fprintf(custom_output_file, "\temruntab = " OPP_CUSTOM_SINK_TAB ";\n");
      else
        fprintf(custom_output_file, "\temruntab = " PUSH_CUSTOM_SINK_TAB ";\n");

      fprintf(custom_output_file, "}\n\n");
    }

    // Print sources
    fprintf(custom_output_file, "#Sources\n");

    for (node_iterator = sources_.begin();
    node_iterator != sources_.end(); node_iterator++){
      node_entry = *node_iterator;

      fprintf(custom_output_file, "node %d {\n", node_entry->node_id_);
      fprintf(custom_output_file, "\tposition = (%.2f, %.2f);\n",
        node_entry->node_x_, node_entry->node_y_);
      if (mode_ == TPP){
        if(send_lambda_ > 0)
          fprintf(custom_output_file, "\temruntab = \"" TPP_CUSTOM_SOURCE_TAB " %.4f\";\n",
            send_lambda_);
        else
          fprintf(custom_output_file, "\temruntab = " TPP_CUSTOM_SOURCE_TAB ";\n");
      }
      else if (mode_ == OPP){
        if(send_lambda_ > 0)
          fprintf(custom_output_file, "\temruntab = \"" OPP_CUSTOM_SOURCE_TAB " %.4f\";\n",
            send_lambda_);
        else
          fprintf(custom_output_file, "\temruntab = " OPP_CUSTOM_SOURCE_TAB ";\n");
      }
      else{
        if(send_lambda_ > 0)
          fprintf(custom_output_file, "\temruntab = \"" PUSH_CUSTOM_SOURCE_TAB " %.4f\";\n",
            send_lambda_);
        else
          fprintf(custom_output_file, "\temruntab = " PUSH_CUSTOM_SOURCE_TAB ";\n");
      }

      fprintf(custom_output_file, "}\n\n");
    }

    // Print other nodes
    fprintf(custom_output_file, "#Intermediate Nodes\n");

    for (node_iterator = nodes_.begin();
    node_iterator != nodes_.end(); node_iterator++){
      node_entry = *node_iterator;

      fprintf(custom_output_file, "node %d {\n", node_entry->node_id_);
      fprintf(custom_output_file, "\tposition = (%.2f, %.2f);\n",
        node_entry->node_x_, node_entry->node_y_);

      fprintf(custom_output_file, "}\n\n");
    }

    // Print Stop time
    fprintf(custom_output_file, "time %dm %ds {\n",
      EXPERIMENT_RUN_MINUTES, EXPERIMENT_RUN_SECONDS);
    fprintf(custom_output_file, "\thalt;\n");
    fprintf(custom_output_file, "}\n");

    if (custom_output_file != stdout)
      fclose(custom_output_file);
  }
  /////////////////////////// Custom Filter //////////////////////////

  if(gen_topo_files_){

    // Output nodes topology file
    sprintf(topology_filename, "nodes-%ld.dat", seed_);
    topology_file = fopen(topology_filename, "w");
    if (topology_file){
      for (node_iterator = nodes_.begin(); node_iterator != nodes_.end(); node_iterator++){
        node_entry = *node_iterator;
        fprintf(topology_file, "%.2f, %.2f\n", node_entry->node_x_, node_entry->node_y_);
      }
      fclose(topology_file);
    }

    // Output sinks topology file
    sprintf(topology_filename, "sinks-%ld.dat", seed_);
    topology_file = fopen(topology_filename, "w");
    if (topology_file){
      for (node_iterator = sinks_.begin(); node_iterator != sinks_.end(); node_iterator++){
        node_entry = *node_iterator;
        fprintf(topology_file, "%.2f, %.2f\n", node_entry->node_x_, node_entry->node_y_);
      }
      fclose(topology_file);
    }

    // Output sources topology file
    sprintf(topology_filename, "sources-%ld.dat", seed_);
    topology_file = fopen(topology_filename, "w");
    if (topology_file){
      for (node_iterator = sources_.begin(); node_iterator != sources_.end(); node_iterator++){
        node_entry = *node_iterator;
        fprintf(topology_file, "%.2f, %.2f\n", node_entry->node_x_, node_entry->node_y_);
      }
      fclose(topology_file);
    }
  }
}

//
// printTrafficFile
//
// This routine outputs the actual ns-2 traffic file.
//
void TopoGen::printTrafficFile()
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;
  char output_filename[100];
  char custom_output_filename[100];
  FILE *output_file, *custom_output_file;
  double start_time = 0.000;
  int tmp_id;

  // Open output files with user specified name if provided
  if (output_filename_){
    strcpy(output_filename, output_filename_);

    // Open a second output file if we are using a custom filter
    if (use_custom_filter_){
      strcpy(custom_output_filename, output_filename_);
      strcat(custom_output_filename, "_");
      strcat(custom_output_filename, CUSTOM_FILTER_NAME);
    }
  }
  // Otherwise automatically generate file names
  else{
    // Changed to include send rate in the file name
    if (mode_ == TPP){
      if (filenames_with_send_rate_)
    sprintf(output_filename, "topo-tpp-%d-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        send_rate_, number_of_runs_);
      else
    sprintf(output_filename, "topo-tpp-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        number_of_runs_);
    }
    else if (mode_ == OPP){
      if (filenames_with_send_rate_)
    sprintf(output_filename, "topo-opp-%d-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        send_rate_, number_of_runs_);
      else
    sprintf(output_filename, "topo-opp-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        number_of_runs_);
    }
    else{
      if (filenames_with_send_rate_)
    sprintf(output_filename, "topo-push-%d-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        send_rate_, number_of_runs_);
      else
    sprintf(output_filename, "topo-push-%d-%d-%d-run-%d",
        number_of_nodes_, number_of_sinks_, number_of_sources_,
        number_of_runs_);
    }

    // Now concantenate the custom string if use_custom_filter
    // Now add the file prefix
    if (use_custom_filter_){
      strcpy(custom_output_filename, output_filename);
      strcat(custom_output_filename, "_");
      strcat(custom_output_filename, CUSTOM_FILTER_NAME);
    }
  }

  strcat(output_filename, ".tcl");
  output_file = fopen(output_filename, "w");
  if (!output_file){
    fprintf(stderr, "Cannot create %s\n", output_filename);
    exit(-1);
  }
  printf("printTrafficFile: opened traffic file: %s\n", output_filename);

  if (use_custom_filter_){
    strcat(custom_output_filename, ".tcl");
    custom_output_file = fopen(custom_output_filename, "w");
    if (!custom_output_file){
      fprintf(stderr, "Cannot create %s\n", custom_output_filename);
      exit(-1);
    }
    printf("printTrafficFile: opened custom traffic file: %s\n", custom_output_filename);
  }

  // Print Configuration Parameters
  fprintf(output_file, "# This is a topo_gen generated ns-2 file\n#\n");
  fprintf(output_file, "# Nodes: %d, Sinks: %d, Sources: %d\n",
    number_of_nodes_, number_of_sinks_, number_of_sources_);
  fprintf(output_file, "# Simulation Area: %d x %d \n",
    size_x_, size_y_ );
  fprintf(output_file, "# Mode: %d, Sinks(%.2f, %.2f), Sources(%.2f, %.2f)\n",
    mode_, a_x_, a_y_, b_x_, b_y_);
  fprintf(output_file, "#Number of nodes = %d;\n",
    number_of_nodes_);
  fprintf(output_file, "#Field size = (%d, %d);\n\n",
    size_x_, size_y_);

  // Print basic ns commands for all nodes
  fprintf(output_file, "# Generate the nodes - the diffusion filter will be added to\n");
  fprintf(output_file, "# each node as specified by node-config in your primary tcl script\n\n");
  for (node_iterator = topology_.begin();
       node_iterator != topology_.end(); node_iterator++){
    node_entry = *node_iterator;

    fprintf(output_file, "$node_(%d) set X_ %f \n", 
        node_entry->node_id_, node_entry->node_x_);
    fprintf(output_file, "$node_(%d) set Y_ %f \n", 
        node_entry->node_id_, node_entry->node_y_);
    fprintf(output_file, "$node_(%d) set Z_ 0.000000\n", node_entry->node_id_);
    fprintf(output_file, "\n");
  }

  // Print sinks
  fprintf(output_file, "#Diffusion Sink Applications\n\n");
  tmp_id = 1;

  for (node_iterator = sinks_.begin();
       node_iterator != sinks_.end(); node_iterator++){
    node_entry = *node_iterator;

    fprintf(output_file, "#Diffusion sink %d application\n", tmp_id);
    fprintf(output_file, "set snk_(%d) [new Application/DiffApp/%sReceiver]\n",
      tmp_id, CUSTOM_FILTER_NAME);
    fprintf(output_file, "$ns_ attach-diffapp $node_(%d) $snk_(%d)\n", 
      node_entry->node_id_, tmp_id);
    fprintf(output_file, "$ns_ at %3.3f \"$snk_(%d) subscribe\"\n\n", start_time, tmp_id);
    tmp_id += 1;
    start_time += .010;
  }

  // Print sources
  fprintf(output_file, "#Diffusion Source Applications\n\n");
  tmp_id = 1;

  for (node_iterator = sources_.begin();
       node_iterator != sources_.end(); node_iterator++){
    node_entry = *node_iterator;
    fprintf(output_file, "#Diffusion source %d application\n", tmp_id);
    fprintf(output_file, "set src_(%d) [new Application/DiffApp/%sSender]\n",
      tmp_id, CUSTOM_FILTER_NAME);
    fprintf(output_file, "$ns_ attach-diffapp $node_(%d) $src_(%d)\n", 
      node_entry->node_id_, tmp_id);
    fprintf(output_file, "$ns_ at %3.3f \"$src_(%d) sourceId %d\"\n",
      start_time, tmp_id, tmp_id);
    start_time += .010;
    fprintf(output_file, "$ns_ at %3.3f \"$src_(%d) publish\"\n\n", start_time, tmp_id);
    tmp_id += 1;
    start_time += .990;
  }
  fclose(output_file);

  ////////////////////// custom filter printfs ////////////////////////
  if (use_custom_filter_){
    // Print Configuration Parameters
    fprintf(custom_output_file, "# This is an ns_topo generated custom filter file\n#\n");
    fprintf(custom_output_file, "# Nodes: %d, Sinks: %d, Sources: %d\n",
      number_of_nodes_, number_of_sinks_, number_of_sources_);
    fprintf(custom_output_file, "# Simulation Area: %d x %d \n",
      size_x_, size_y_ );
    fprintf(custom_output_file, "# Mode: %d, Sinks(%.2f, %.2f), Sources(%.2f, %.2f)\n",
      mode_, a_x_, a_y_, b_x_, b_y_);
    fprintf(custom_output_file, "#Number of nodes = %d;\n",
      number_of_nodes_);
    fprintf(custom_output_file, "#Field size = (%d, %d);\n\n",
      size_x_, size_y_);

    // Print basic ns commands for all nodes
    fprintf(output_file, "# Generate the nodes - the diffusion filter will be added to\n");
    fprintf(output_file, "#   each node as specified by node-config in your primary tcl script\n\n");
    for (node_iterator = topology_.begin();
         node_iterator != topology_.end(); node_iterator++){
      node_entry = *node_iterator;

      fprintf(custom_output_file, "$node_(%d) set X_ %f \n", 
          node_entry->node_id_, node_entry->node_x_);
      fprintf(custom_output_file, "$node_(%d) set Y_ %f \n", 
          node_entry->node_id_, node_entry->node_y_);
      fprintf(custom_output_file, "$node_(%d) set Z_ 0.000000\n", node_entry->node_id_);
      fprintf(custom_output_file, "\n");
    }

    fprintf (custom_output_file, "# Add a Custom Filter to each node\n\n");

    for (node_iterator = topology_.begin();
         node_iterator != topology_.end(); node_iterator++){
      node_entry = *node_iterator;

      fprintf(custom_output_file, "#Add a Custom Filter to node %d;\n",
        node_entry->node_id_);
      fprintf(custom_output_file, "set flt_(%d) [new Application/DiffApp/%sFilter]\n", 
        node_entry->node_id_, CUSTOM_FILTER_NAME);
      fprintf(custom_output_file, "$ns_ attach-diffapp $node_(%d) $flt_(%d)\n", 
        node_entry->node_id_, node_entry->node_id_);
      fprintf(custom_output_file, "$ns_ at %3.3f \"$flt_(%d) start\"\n\n",
        start_time, node_entry->node_id_);
      start_time += .010;
    }

    // Print sinks
    fprintf(custom_output_file, "#Diffusion Sink Applications\n\n");
    tmp_id = 1;

    for (node_iterator = sinks_.begin();
         node_iterator != sinks_.end(); node_iterator++){
      node_entry = *node_iterator;

      fprintf(custom_output_file, "#Diffusion sink %d application\n", tmp_id);
      fprintf(custom_output_file, "set snk_(%d) [new Application/DiffApp/%sReceiver]\n",
        tmp_id, CUSTOM_FILTER_NAME);
      fprintf(custom_output_file, "$ns_ attach-diffapp $node_(%d) $snk_(%d)\n",
        node_entry->node_id_, tmp_id);
      fprintf(custom_output_file, "$ns_ at %3.3f \"$snk_(%d) subscribe\"\n\n",
        start_time, tmp_id);
      tmp_id += 1;
      start_time += .010;
    }

    // Print sources
    fprintf(custom_output_file, "#Diffusion Source Applications\n\n");
    tmp_id = 1;

    for (node_iterator = sources_.begin();
         node_iterator != sources_.end(); node_iterator++){
      node_entry = *node_iterator;
      fprintf(custom_output_file, "#Diffusion source %d application\n", tmp_id);
      fprintf(custom_output_file, "set src_(%d) [new Application/DiffApp/%sSender]\n",
        tmp_id, CUSTOM_FILTER_NAME);
      fprintf(custom_output_file, "$ns_ attach-diffapp $node_(%d) $src_(%d)\n",
        node_entry->node_id_, tmp_id);
      fprintf(custom_output_file, "$ns_ at %3.3f \"$src_(%d) sourceId %d\"\n",
        start_time, tmp_id, tmp_id);
      start_time += .010;
      fprintf(custom_output_file, "$ns_ at %3.3f \"$src_(%d) publish\"\n\n",
        start_time, tmp_id);
      tmp_id += 1;
      start_time += .990;
    }
    fclose(custom_output_file);
  }
  printf("printTrafficFile: done!\n");
}

void TopoGen::statsConfig()
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;
  int topology_diameter;
  int current_diameter;

  fprintf(stdout, "Checking topology...\n");
  topology_diameter = 0;

  // Go through all sinks
  for (node_iterator = sinks_.begin();
       node_iterator != sinks_.end(); node_iterator++){
    node_entry = *node_iterator;

    current_diameter = statsNode(node_entry->node_id_);

    if (current_diameter > topology_diameter)
      topology_diameter = current_diameter;
  }
  fprintf(stdout, "Network diameter is : %d\n", topology_diameter);
}

int TopoGen::statsNode(int32_t node_id)
{
  NodeList::iterator node_iterator;
  TopoList::iterator topo_list_itr;
  CostList::iterator cost_list_itr;
  TopoList *to_process;
  TopoList *in_processing;
  CostList processed;
  NodeInfo *node_entry;
  CostInfo *cost_entry;
  int current_diameter, max_diameter, current_cost;
  bool already_processed;
  double node_x, node_y;
  int32_t current_node;

  // Insert initial node in the to_process list
  to_process = new TopoList;
  to_process->push_back(node_id);
  current_cost = 0;

  // Start node has cost 0
  cost_entry = new CostInfo(node_id, 0);
  processed.push_back(cost_entry);

  // Process topology
  while(to_process->size()){

    // Move nodes from to_porcess to in_processing
    in_processing = to_process;
    to_process = new TopoList;

    // Increase cost
    current_cost++;

    // Process all nodes in the in_processing list
    for (topo_list_itr = in_processing->begin();
      topo_list_itr != in_processing->end();){

      // Get node and delete it from the list
      current_node = *topo_list_itr;
      topo_list_itr = in_processing->erase(topo_list_itr);

      node_entry = findNode(&topology_, current_node);

      if (!node_entry){
        fprintf(stderr, "Cannot find node !\n");
        exit(-1);
      }
      
      node_x = node_entry->node_x_;
      node_y = node_entry->node_y_;

      for (node_iterator = topology_.begin();
       node_iterator != topology_.end(); node_iterator++){

        node_entry = *node_iterator;

        // Check if the node is within radio range of node_id
        if (Distance(node_x, node_y,
             node_entry->node_x_,
             node_entry->node_y_) < COMMUNICATION_RANGE){

          // We can reach this guy, check if it's already in the CostList
          already_processed = false;

          for (cost_list_itr = processed.begin();
            cost_list_itr != processed.end(); cost_list_itr++){
            cost_entry = *cost_list_itr;

            // Compare node ids
            if (cost_entry->node_id_ == node_entry->node_id_){
              already_processed = true;
              break;
            }
          }

          if (!already_processed){
            // Insert this node in the processed and to_process lists
            to_process->push_back(node_entry->node_id_);
            cost_entry = new CostInfo(node_entry->node_id_, current_cost);
            processed.push_back(cost_entry);
          }
        } // if Distance...
      } // for node_iterator
    } // for topo_list_itr

    if (in_processing->size()){
      fprintf(stderr, "in_processing list not empty... size %d !\n",
        in_processing->size());
      exit(-1);
    }

    in_processing->clear();
    delete in_processing;
  } // while loop

  // Clean up
  to_process->clear();
  delete to_process;

  max_diameter = 0;

  for (node_iterator = sources_.begin();
       node_iterator != sources_.end(); node_iterator++){

    // Check source's cost
    node_entry = *node_iterator;
    current_diameter = 0;

    for (cost_list_itr = processed.begin(); cost_list_itr != processed.end();
     cost_list_itr++){
      cost_entry = *cost_list_itr;

      if (cost_entry->node_id_ == node_entry->node_id_){
        current_diameter = cost_entry->cost_;
        break;
      }
    }

    if (current_diameter == 0){
      fprintf(stderr, "Source %d couldn't be found !\n", node_entry->node_id_);
      exit(-1);
    }

    if (max_diameter < current_diameter)
      max_diameter = current_diameter;
  }

  // Done, clean up now
  for (cost_list_itr = processed.begin(); cost_list_itr != processed.end();
       cost_list_itr++){
    delete (*cost_list_itr);
  }
  processed.clear();

  return max_diameter;;
}

bool TopoGen::checkConfig()
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;
  int start_node, node_id;

  // Pick one source as our start node
  start_node = (*(sources_.begin()))->node_id_;

  // Node can connect to itself, so we add it to the connected list
  connected_.push_back(start_node);

  // Build list of connected nodes
  checkConnected(start_node);

  // All sinks and sources must be connected

  // Start with sinks
  for (node_iterator = sinks_.begin();
       node_iterator != sinks_.end(); node_iterator++){
    node_entry = *node_iterator;

    if (!isConnected(&connected_, node_entry->node_id_)){
      return false;
    }
  }

  // Now check sources
  for (node_iterator = sources_.begin();
       node_iterator != sources_.end(); node_iterator++){
    node_entry = *node_iterator;

    if (!isConnected(&connected_, node_entry->node_id_)){
      return false;
    }
  }

  return true;
}

void TopoGen::checkConnected(int node_id)
{
  NodeList::iterator node_iterator;
  NodeInfo *node_entry;
  double node_x, node_y;

  node_entry = findNode(&topology_, node_id);

  if (!node_entry){
    fprintf(stderr, "Cannot find node !\n");
    exit(-1);
  }

  node_x = node_entry->node_x_;
  node_y = node_entry->node_y_;

  for (node_iterator = topology_.begin();
       node_iterator != topology_.end(); node_iterator++){

    node_entry = *node_iterator;

    // Check if the node is within radio range of node_id
    if (Distance(node_x, node_y,
         node_entry->node_x_,
         node_entry->node_y_) < COMMUNICATION_RANGE){
      // We can reach this guy
      if (!isConnected(&connected_, node_entry->node_id_)){
    // This node is connected, add it to the list
    connected_.push_back(node_entry->node_id_);

    // Repeat this for its neighbors
    checkConnected(node_entry->node_id_);
      }
    }
  }
}

bool TopoGen::isConnected(TopoList *topology_list, int node_id)
{
  TopoList::iterator node_iterator;

  for (node_iterator = topology_list->begin();
       node_iterator != topology_list->end(); node_iterator++){
    
    if (*node_iterator == node_id)
      return true;
  }

  // Not found
  return false;
}

void TopoGen::doIt()
{
  struct timeval tmv;

  while(number_of_runs_ > 0){

    // Initialize
    initialize();
    fprintf(stderr, "Starting run %d, using seed %d\n",
        number_of_runs_, seed_);

    // Create all nodes
    createNodes();

    // Pick locations a and b in the region
    if (position_ & DIAGONAL){
      // If the user wants diagonally opposite sources and sinks
      // a should be in the lower-left corner
      a_x_ = 0.0;
      a_y_ = 0.0;
      // b should be in the upper-right corner
      b_x_ = size_x_;
      b_y_ = size_y_;
    }
    else{
      // a should be in the left-lower quadrant
      a_x_ = 1.0 * size_x_ / 4 * (GetRand() * 1.0 / RAND_MAX);
      a_y_ = 1.0 * size_y_ / 4 * (GetRand() * 1.0 / RAND_MAX);

      // b must be in the right-upper quadrant
      b_x_ = 1.0 * size_x_ / 4 * (GetRand() * 1.0 / RAND_MAX);
      b_y_ = 1.0 * size_y_ / 4 * (GetRand() * 1.0 / RAND_MAX);

      // Move it up, right
      b_x_ = b_x_ + (size_x_ * 3 / 4);
      b_y_ = b_y_ + (size_y_ * 3 / 4);
    }

    if (position_ & CLUSTERED_SINKS){
      if (!pickClusteredSinks(a_x_, a_y_)){
        fprintf(stderr, "Error, Cannot pick clustered sinks !\n");
        exit(-1);
      }
    }
    else{
      if (!pickRandomSinks()){
        fprintf(stderr, "Error, Cannot pick random sinks !\n");
        exit(-1);
      }
    }

    if (position_ & CLUSTERED_SOURCES){
      if (!pickClusteredSources(b_x_, b_y_)){
        fprintf(stderr, "Error, Cannot pick clustered sources !\n");
        exit(-1);
      }
    }
    else{
      if (!pickRandomSources()){
        fprintf(stderr, "Error, Cannot pick random sources !\n");
        exit(-1);
      }
    }


    // Check if topology is usable
    if (checkConfig()){

      if (test_only_){
    // We only check the topology and print out information
    statsConfig();
      }
      else{
        // Yes, generate configuration files for this topology
        if(gen_ns2_files_)
          printTrafficFile();
        else
      printConfig();
      }

      // Move to other runs
      number_of_runs_--;
    }

    // Increment seed
    seed_++;
  }

  // We are done !
  GetTime(&tmv);
  fprintf(stderr, "Done at %ld.%06ld\n", tmv.tv_sec, tmv.tv_usec);
}

void TopoGen::initialize()
{
  nodes_.clear();
  sinks_.clear();
  sources_.clear();
  topology_.clear();
  connected_.clear();
}

TopoGen::TopoGen(int argc, char **argv)
{
  struct timeval tmv;

  // Initialize default values
  size_x_ = 0;
  size_y_ = 0;
  mode_ = 0;
  position_ = RANDOM;
  user_position_ = 1;
  number_of_sinks_ = 1;
  number_of_sources_ = 1;
  number_of_nodes_ = 2;
  output_filename_ = NULL;
  seed_ = 0;
  send_rate_ = 0;
  send_lambda_ = 0;
  number_of_runs_ = 1;
  use_custom_filter_ = false;
  equalize_events_ = true;
  gen_topo_files_ = false;
  gen_ns2_files_ = false;
  test_only_ = false;
  filenames_with_send_rate_ = false;

  // Parse command line options
  parseCommandLine(argc, argv);

  // Initialize Topology Generator
  GetTime(&tmv);

  if (seed_ == 0){
    SetSeed(&tmv);
    seed_ = tmv.tv_usec;
  }
  else
    srand(seed_);
  
  if (send_rate_ > 0){
    // Calculate send_lambda for the poisson model on how often source
    // nodes send events. If we want to send the same number of events
    // when having different number of sources in the network,
    // send_lambda_ has to be divided by the number of sources in order
    // to have the same overall number of events.
    if (equalize_events_)
      send_lambda_ = (float) 1.0 / (number_of_sources_ * send_rate_);
    else
      send_lambda_ = (float) 1.0 / send_rate_;
  }

  fprintf(stderr, "Initializing at %ld.%06ld\n",
      tmv.tv_sec, tmv.tv_usec);
}

void TopoGen::parseCommandLine(int argc, char **argv)
{
  int opt;
  
  opterr = 0;

  while (1){
    opt = getopt(argc, argv, "htegicdza:b:m:x:y:f:n:s:r:k:p:");

    switch (opt){

    case 'x':
      size_x_ = atoi(optarg);
      if (size_x_ < MIN_X || size_x_ > MAX_X){
        fprintf(stderr, "Error: Invalid X size !\n");
        usage(argv[0]);
      }
      break;

    case 'y':
      size_y_ = atoi(optarg);
      if (size_y_ < MIN_Y || size_y_ > MAX_Y){
        fprintf(stderr, "Error: Invalid Y size !\n");
        usage(argv[0]);
      }
      break;

    case 'n':
      number_of_nodes_ = atoi(optarg);
      if (number_of_nodes_ < MIN_NODES || number_of_nodes_ > MAX_NODES){
        fprintf(stderr, "Error: Invalid number of nodes !\n");
        usage(argv[0]);
      }
      break;

    case 's':
      seed_ = atoi(optarg);
      if (seed_ < 0){
        fprintf(stderr, "Error: Invalid seed !\n");
        usage(argv[0]);
      }
      break;

    case 'r':
      number_of_runs_ = atoi(optarg);
      if (number_of_runs_ < 1){
        fprintf(stderr, "Error: Number of runs must be > 0 !\n");
        usage(argv[0]);
      }
      break;

    case 'a':
      number_of_sinks_ = atoi(optarg);
      if (number_of_sinks_ < MIN_SINKS || number_of_sinks_ > MAX_SINKS){
        fprintf(stderr, "Error: Invalid number of sinks !\n");
        usage(argv[0]);
      }
      break;

    case 'b':
      number_of_sources_ = atoi(optarg);
      if (number_of_sources_ < MIN_SOURCES || number_of_sources_ > MAX_SOURCES){
        fprintf(stderr, "Error: Invalid number of sources !\n");
        usage(argv[0]);
      }
      break;

    case 'm':
      mode_ = atoi(optarg);
      if (mode_ < MIN_MODE || mode_ > MAX_MODE){
        fprintf(stderr, "Error: Invalid mode !\n");
        usage(argv[0]);
      }
      break;

    case 'p':
      user_position_ = atoi(optarg);
      if (user_position_ < MIN_POS || user_position_ > MAX_POS){
        fprintf(stderr, "Error: Invalid position !\n");
        usage(argv[0]);
      }
      if (user_position_ == 1)
        position_ = RANDOM;
      else if (user_position_ == 2)
        position_ = CLUSTERED_SINKS;
      else if (user_position_ == 3)
        position_ = CLUSTERED_SOURCES;
      else if (user_position_ == 4)
        position_ = CLUSTERED_SINKS|CLUSTERED_SOURCES;
      else if (user_position_ == 5)
        position_ = DIAGONAL|CLUSTERED_SINKS|CLUSTERED_SOURCES;
      break;

    case 'k':
      send_rate_ = atoi(optarg);
      if (send_rate_ < MIN_SEND_RATE || send_rate_ > MAX_SEND_RATE){
        fprintf(stderr, "Error: Invalid send rate !\n");
        usage(argv[0]);
      }
      break;

    case 'f':
      if (!strncasecmp(optarg, "-", 1)){
        fprintf(stderr, "Error: Parameter is missing !\n");
        usage(argv[0]);
      }
      output_filename_ = strdup(optarg);
      break;

    case 'c':
      use_custom_filter_ = true;
      break;

    case 'e':
      equalize_events_ = false;
      break;

    case 'g':
      gen_topo_files_ = true;
      break;

    case 't':
      test_only_ = true;
      break;

    case 'i':
      filenames_with_send_rate_ = true;
      break;

    case 'h':
      usage(argv[0]);
      break;

    case 'z':
      gen_ns2_files_ = true;
      break;

    case '?':
      fprintf(stderr,
          "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage(argv[0]);
      break;

    case ':':
      fprintf(stderr, "Error: Parameter missing !\n");
      usage(argv[0]);
      break;
    }

    if (opt == -1)
      break;
  }

  if (size_x_ == 0 || size_y_ == 0 || mode_ == 0 ||
      number_of_sinks_ == 0 || number_of_sources_ == 0 ||
      number_of_nodes_ == 0)
    usage(argv[0]);

  if (output_filename_ && number_of_runs_ > 1)
    usage(argv[0]);
}

void TopoGen::usage(char *s)
{
  fprintf(stderr, "Usage: %s -x x_size -y y_size -m mode [-n #nodes] [-a #sinks] [-b #sources] [-p position] [-r #runs] [-c] [-f outputfile] [-s seed] [-k sendrate] [-i] [-e] [-g] [-t] [-h] [-z]\n\n", s);
  fprintf(stderr, "\t-x - Sets region X size\n");
  fprintf(stderr, "\t-y - Sets region Y size\n");
  fprintf(stderr, "\t-m - Selects mode: 1 - Push Diffusion\n");
  fprintf(stderr, "\t                   2 - OnePhasePull Diffusion\n");
  fprintf(stderr, "\t                   3 - TwoPhasePull Diffusion\n");
  fprintf(stderr, "\t-n - Selects number of nodes (default=2)\n");
  fprintf(stderr, "\t-a - Selects number of sinks (default=1)\n");
  fprintf(stderr, "\t-b - Selects number of sources (default=1)\n");
  fprintf(stderr, "\t-p - Selects Position: 1 - Random Sources and Sinks (default)\n");
  fprintf(stderr, "\t                       2 - Clustered Sinks\n");
  fprintf(stderr, "\t                       3 - Clustered Sources\n");
  fprintf(stderr, "\t                       4 - Clustered Sources and Sinks\n");
  fprintf(stderr, "\t                       5 - Diagonally Opposite Sources and Sinks\n");
  fprintf(stderr, "\t-r - Selects number of runs (default=1)\n");
  fprintf(stderr, "\t-c - Add a custom filter\n");
  fprintf(stderr, "\t-f - Selects output file (default=autogenerated)\n");
  fprintf(stderr, "\t-s - Sets the random number generation seed (default=TOD)\n");
  fprintf(stderr, "\t-k - Selects send rate in seconds (default=none specified)\n");
  fprintf(stderr, "\t-i - Writes filenames with send rate information\n");
  fprintf(stderr, "\t-e - Disables event equalization (default=enabled)\n");
  fprintf(stderr, "\t-g - Enables generation of topology position files\n");
  fprintf(stderr, "\t-t - Generates/Tests the topology (doesn't output config file)\n");
  fprintf(stderr, "\t-h - Prints this information\n");
  fprintf(stderr, "\t-z - Output ns-2 files (default is EmSim)\n");
  fprintf(stderr, "\n");
  exit(0);
}

int main(int argc, char **argv)
{
  TopoGen *topo;

  topo = new TopoGen(argc, argv);
  topo->doIt();

  return 0;
}
